using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;

//OpsLib is the namespace that has the highest-level API functions.
using VaultClientOperationsLib;
//ClientService is the namespace for all of the types that travel between the client and the server.
using VaultClientNetLib.ClientService; 
using System.Collections;

namespace FindFiles
{
	class VaultFindFiles
	{
		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			if ( args.Length < 4 || args.Length > 9 )
			{
				usage();
				return;
			}

			string hostName, username, password, repository, searchRoot, filename;
			bool exactMatch, includeDate, includeVersion, caseInsensitive;

			try
			{
				ParseArguments(args, out exactMatch, out includeDate, out includeVersion, 
								out caseInsensitive, out hostName, out repository, out username, 
								out password, out searchRoot, out filename);
			}
			catch(ArgumentException)
			{
				usage();
				return;
			}
			
			ClientInstance myClient = new ClientInstance();
			myClient.Init(VaultClientNetLib.VaultConnection.AccessLevelType.Client);
			
			try
			{
				myClient.Login(hostName, username, password);
			}
			catch(System.Web.Services.Protocols.SoapException)
			{
				Console.WriteLine("Error logging into Vault with username: " + username);
				return;
			}

			try
			{
				FindRepository(myClient, repository);
			}
			catch(ArgumentException)
			{
				Console.WriteLine("Unable to locate repository: " + repository);
				return;
			}

			try
			{
				VaultClientFolder searchFolder = myClient.Repository.Root.FindFolderRecursive(searchRoot);

				if ( searchFolder == null )
				{
					Console.WriteLine("Didn't find " + filename + " under " + searchRoot + ".");
					return;
				}

				ArrayList returnArray = new ArrayList();
				RecursivelySearchForFileName(searchFolder, filename, ref returnArray, 
												exactMatch, caseInsensitive);
				
				if ( returnArray.Count == 0 )
				{
					Console.WriteLine("Didn't find " + filename + " under " + searchRoot + ".");
					return;
				}

				foreach(Object anObj in returnArray)
				{
					if (anObj is VaultClientFile)
					{
						VaultClientFile theFile = (VaultClientFile)anObj;
						StringBuilder outText = new StringBuilder(theFile.FullPath);
						if ( includeDate )
							outText.Append(" " + theFile.ModifiedDate);
			
						if ( includeVersion )
							outText.Append(" Version: " + theFile.Version);

						Console.WriteLine(outText + Environment.NewLine);
					}
				}

				Console.WriteLine("Total files found: " + returnArray.Count.ToString());
			}
			catch(System.Web.Services.Protocols.SoapException)
			{
				Console.WriteLine("Error searching folders.");
				return;
			}
		}
		
		public static void RecursivelySearchForFileName(VaultClientFolder folder, string fileName, 
														ref ArrayList returnArray, bool exactMatch,
														bool caseInsensitive)
		{
			if (folder.Files != null)
			{
				foreach (VaultClientFile subfile in folder.Files)
				{
					if ( exactMatch )
					{
						if ( caseInsensitive && subfile.Name.ToUpper() == fileName.ToUpper() )
						{
							returnArray.Add(subfile);
						}
						else if ( subfile.Name == fileName )
						{
							returnArray.Add(subfile);
						}
					}
					else
					{
						if ( caseInsensitive && 
								Regex.IsMatch(subfile.Name, fileName, RegexOptions.IgnoreCase) )
						{
							returnArray.Add(subfile);
						}
						else if ( Regex.IsMatch(subfile.Name, fileName) )
						{
							returnArray.Add(subfile);
						}
					}
				}
			}
			
			if (folder.Folders != null)
			{
				foreach (VaultClientFolder subfolder in folder.Folders)
				{
					RecursivelySearchForFileName(subfolder, fileName, ref returnArray, exactMatch,
												caseInsensitive);
				}
			}
		}

		/// <summary>
		/// This is a helper function that works around a deficiency in the 
		/// Vault Client API.  In the API, there is no method that lets you set the active repository
		/// for a ClientInstance based only on the repository name.  This means that (for now) every
		/// program that uses the API will need to do something like this.
		/// </summary>
		/// <param name="client">A Client Instance that has already had Init and Login called on it.</param>
		/// <param name="repositoryName">The name of the repository to find and connect to.</param>
		private static void FindRepository(ClientInstance client, string repositoryName)
		{
			VaultRepositoryInfo[] reps = null;
			//List all the repositories on the server.
			client.ListRepositories(ref reps);

			//Search for the one that we want.
			foreach (VaultRepositoryInfo r in reps)
			{
				if (String.Compare(r.RepName,repositoryName, true) == 0)
				{
					//This will load up the client side cache files and refresh the repository structure.
					//See http://support.sourcegear.com/viewtopic.php?t=6 for more on client side cache files.
					client.SetActiveRepositoryID(r.RepID, client.Connection.Username, r.UniqueRepID, true, true);
					break;
				}
			}
			if (client.ActiveRepositoryID == -1)
				throw new ArgumentException(string.Format("Repository {0} not found", repositoryName));
		}

		private static void ParseArguments(string[] args, out bool exactMatch, out bool includeDate, 
											out bool includeVersion, out bool caseInsensitive,
											out string hostName, out string respository, out string username, 
											out string password, out string searchRoot, out string filename)
		{
			// defaults
			exactMatch = false;
			includeDate = false;
			includeVersion = false;
			caseInsensitive = false;
			hostName = "";
			respository = "";
			username = "";
			password = "";
			searchRoot = "";
			filename = "";

			foreach ( string anArg in args )
			{
				switch ( anArg.ToUpper().Substring(0, 2) )
				{
					case "-E":
						if ( anArg.Trim().Length > 2 )
							throw new ArgumentException();
						exactMatch = true;
						break;
					case "-D":
						if ( anArg.Trim().Length > 2 )
							throw new ArgumentException();
						includeDate = true;
						break;
					case "-V":
						if ( anArg.Trim().Length > 2 )
							throw new ArgumentException();
						includeVersion = true;
						break;
					case "-I":
						if ( anArg.Trim().Length > 2 )
							throw new ArgumentException();
						caseInsensitive = true;
						break;
					case "-H":
						hostName = anArg.Substring(2);
						if (hostName.StartsWith("http://") == false)
							hostName = "http://" + hostName;
						if (hostName.EndsWith("/VaultService") == false)
							hostName += "/VaultService";
						break;
					case "-R":
						respository = anArg.Substring(2);
						break;
					case "-U":
						username = anArg.Substring(2);
						break;
					case "-P":
						password = anArg.Substring(2);
						break;
					case "-S":
						searchRoot = anArg.Substring(2);
						if ( searchRoot.Substring(searchRoot.Length - 1) == "/" )
							searchRoot = searchRoot.Substring(0, searchRoot.Length -1);
						break;
					case "-F":
						filename = anArg.Substring(2);
						break;
					default:
						break;
						// do nothing
				}
			}

			// default searchRoot if not specified
			if ( searchRoot.Length == 0 )
				searchRoot = "$";

			// default user to Windows logon if not supplied
			if ( username.Length == 0 )
			{
				username = System.Security.Principal.WindowsIdentity.GetCurrent().Name;
				// remove domain if necessary
				if ( username.IndexOf('\\') != -1 )
				{
					username = username.Substring( username.IndexOf('\\') + 1 );
				}
			}
		}

		private static void usage()
		{
			Console.WriteLine("FindFiles - locates files by name within a Vault repository.");
			Console.WriteLine("usage: " +
				"FindFiles [OPTIONS] -h[vault host] -r[repository] -u[username] -p[password] " +
				"-s[search root] -f[filename]");
			Console.WriteLine(Environment.NewLine + "OPTIONS: ");
			Console.WriteLine("-r Vault repository to search (required)");
			Console.WriteLine("-s Search root - location in Vault repository to start search (default: /)");
			Console.WriteLine("-u Username - Vault username to use to log on (default: current Windows username).");
			Console.WriteLine("-p Password for Vault user (required)");
			Console.WriteLine("-i Search is case insensitive (default: case sensitive)");
			Console.WriteLine("-e Exact match - no wildcards used for filename search " + 
									"(default: wildcard match with regular expressions).");
			Console.WriteLine("-d Include last modified date in output (default: no).");
			Console.WriteLine("-v Include version number for each file found (default: no).");
			Console.WriteLine(Environment.NewLine + "For example:");
			Console.WriteLine("FindFiles -e -hvault.hrbl.net -rGriffin -usomeuser -ppwd123 -s$/System -fAssemblyInfo.vb");
			Console.WriteLine("FindFiles -hvault.hrbl.net -rGriffin -ppwd123 -f.*.cs$");
		}
	}
}
